/* -*- c-basic-offset: 3 -*-
 *
 * PANDORE (PANTHEON Project)
 *
 * GREYC IMAGE
 * 6 Boulevard Marchal Juin
 * F-14050 Caen Cedex France
 *
 * This file is free software. You can use it, distribute it
 * and/or modify it. However, the entire risk to the quality
 * and performance of this program is with you.
 *
 * Please email any bugs, comments, and/or additions to this file to:
 * pandore@greyc.ensicaen.fr
 * http://www.greyc.ensicaen.fr/EquipeImage/Pandore/
 */

/**
 * @author Jean-Luc Starck (CEA 1996).
 */

#ifndef _FITSIO_H
#define _FITSIO_H

#include <stdio.h>
/* stddef.h is apparently needed to define size_t */
#include <stddef.h>
 
/* global variables */ 
#define FLEN_FILENAME 257  /* max length of a filename  */
#define FLEN_KEYWORD    9  /* max length of a keyword */
#define FLEN_CARD      81  /* length of a FITS header card */
#define FLEN_VALUE     71  /* max length of a keyword value string */
#define FLEN_COMMENT   73  /* max length of a keyword comment string */
#define FLEN_ERRMSG    81  /* max length of a FITSIO error message */
#define FLEN_STATUS    31  /* max length of a FITSIO status text string */
 
#define TBIT          1  /* codes for FITS table data types */
#define TBYTE        11
#define TLOGICAL     14
#define TSTRING      16
#define TUSHORT      20
#define TSHORT       21
#define TUINT        30
#define TINT         31
#define TULONG       40
#define TLONG        41
#define TFLOAT       42
#define TDOUBLE      82
#define TCOMPLEX     83
#define TDBLCOMPLEX 163

#define INT32BIT int  /* 32-bit integer datatype.  Currently this       */
                      /* datatype is an 'int' on all useful platforms   */
                      /* however, it is possible that that are cases    */
                      /* where 'int' is a 2-byte integer, in which case */
                      /* FITSINT would need to be defined as 'long'.    */

#define BYTE_IMG      8  /* BITPIX code values for FITS image types */
#define SHORT_IMG    16
#define LONG_IMG     32
#define FLOAT_IMG   -32
#define DOUBLE_IMG  -64
                         /* The following 2 codes are not true FITS         */
                         /* datatypes; these codes are only used internally */
                         /* within cfitsio to make it easier for users      */
                         /* to deal with unsigned integers.                 */
#define USHORT_IMG   20
#define ULONG_IMG    40

#define IMAGE_HDU  0  /* Primary Array or IMAGE HDU */
#define ASCII_TBL  1  /* ASCII table HDU  */
#define BINARY_TBL 2  /* Binary table HDU */
#define ANY_HDU   -1  /* matches any HDU type */

#define READONLY  0    /* options when opening a file */
#define READWRITE 1

/* adopt a hopefully obscure number to use as a null value flag */
/* could be problems if the FITS files contain data with these values */
#define FLOATNULLVALUE -9.11912E-36F
#define DOUBLENULLVALUE -9.1191291391491E-36
 
#ifndef TRUE
#define TRUE 1
#else
#undef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define CASESEN   1   /* do case-sensitive string match */
#define CASEINSEN 0   /* do case-insensitive string match */
 
#define MAXHDU 1000    /* maximum number of extensions allowed in a FITS file */

#define GT_ID_ALL_URI  0   /* hierarchical grouping parameters */
#define GT_ID_REF      1
#define GT_ID_POS      2
#define GT_ID_ALL      3
#define GT_ID_REF_URI 11
#define GT_ID_POS_URI 12

#define OPT_RM_GPT      0
#define OPT_RM_ENTRY    1
#define OPT_RM_MBR      2
#define OPT_RM_ALL      3

#define OPT_GCP_GPT     0
#define OPT_GCP_MBR     1
#define OPT_GCP_ALL     2

#define OPT_MCP_ADD     0
#define OPT_MCP_NADD    1
#define OPT_MCP_REPL    2
#define OPT_MCP_MOV     3

#define OPT_MRG_COPY    0
#define OPT_MRG_MOV     1

#define OPT_CMT_MBR      1
#define OPT_CMT_MBR_DEL 11

/* structure used to store table column information */
typedef struct {
    char ttype[70];   /* column name = FITS TTYPEn keyword; */
    long tbcol;       /* offset in row to first byte of each column */
    int  tdatatype;   /* datatype code of each column */
    long  trepeat;     /* repeat count of column; number of elements */
    double tscale;    /* FITS TSCALn linear scaling factor */
    double tzero;     /* FITS TZEROn linear scaling zero point */
    long tnull;       /* FITS null value for int image or binary table cols */
    char strnull[20]; /* FITS null value string for ASCII table columns */
    char tform[10];   /* FITS tform keyword value  */
    long  twidth;     /* width of each ASCII table column */
}tcolumn;

#define VALIDSTRUC 555  /* magic value used to identify if structure is valid */

/* structure used to store basic FITS file information */
typedef struct {
    int filehandle; /* handle returned by the file open function */
    int driver;     /* defines which set of I/O drivers should be used */
    int open_count; /* number of opened 'fitsfiles' using this structure */
    char *filename; /* file name */
    int validcode;  /* magic value used to verify that structure is valid */
    long filesize;  /* current size of the physical disk file in bytes */
    long logfilesize; /* logical size of file, including unflushed buffers */
    int lasthdu;    /* is this the last HDU in the file? 0 = no, else yes */
    long bytepos;   /* current logical I/O pointer position in file */
    long io_pos;    /* current I/O pointer position in the physical file */
    int curbuf;     /* number of I/O buffer currently in use */ 
    int curhdu;     /* current HDU number; 0 = primary array */
    int hdutype;    /* 0 = primary array, 1 = ASCII table, 2 = binary table */
    int writemode;  /* 0 = readonly, 1 = readwrite */
    int maxhdu;     /* highest numbered HDU known to exist in the file */
    long headstart[MAXHDU + 1]; /* byte offset in file to start of each HDU */
    long headend;   /* byte offest in file to end of the current HDU header */
    long nextkey;   /* byte offset in file to beginning of next keyword */
    long datastart; /* byte offset in file to start of the current data unit */
    int tfield;     /* number of fields in the table (primary array has 2 */
    long origrows;  /* original number of rows (value of NAXIS2 keyword)  */
    long numrows;   /* number of rows in the table (dynamically updated) */
    long rowlength; /* total length of a table row, in bytes */
    tcolumn *tableptr; /* pointer to the table structure */
    long heapstart; /* heap start byte relative to start of data unit */
    long heapsize;  /* size of the heap, in bytes */
} FITSfile;

/* structure used to store basic HDU information */
typedef struct {
    int HDUposition;  /* HDU position in file; 0 = first HDU */
    FITSfile *Fptr;   /* pointer to FITS file structure */
}fitsfile;

/* structure for the iterator function column information */
typedef struct {  
     /* elements required as input to fits_iterate_data: */

    fitsfile *fptr;     /* pointer to the HDU containing the column */
    int      colnum;    /* column number in the table (use name if < 1) */
    char     colname[70]; /* name (= TTYPEn value) of the column (optional) */
    int      datatype;  /* output datatype (converted if necessary  */
    int      iotype;    /* = InputCol, InputOutputCol, or OutputCol */

    /* output elements that may be useful for the work function: */

    void     *array;    /* pointer to the array (and the null value) */
    long     repeat;    /* binary table vector repeat value */
    long     tlmin;     /* legal minimum data value */
    long     tlmax;     /* legal maximum data value */
    char     tunit[70]; /* physical unit string */
    char     tdisp[70]; /* suggested display format */

} iteratorCol;

#define InputCol         0  /* flag for input only iterator column       */
#define InputOutputCol   1  /* flag for input and output iterator column */
#define OutputCol        2  /* flag for output only iterator column      */

/* error status codes */
 
#define USE_MEM_BUFF     -101  /* use memory buffer when opening file */
#define OVERFLOW_ERR      -11  /* overflow during datatype conversion */
#define SAME_FILE         101  /* input and output files are the same */
#define TOO_MANY_FILES    103  /* tried to open too many FITS files */
#define FILE_NOT_OPENED   104  /* could not open the named file */
#define FILE_NOT_CREATED  105  /* could not create the named file */
#define WRITE_ERROR       106  /* error writing to FITS file */
#define END_OF_FILE       107  /* tried to move past end of file */
#define READ_ERROR        108  /* error reading from FITS file */
#define FILE_NOT_CLOSED   110  /* could not close the file */
#define ARRAY_TOO_BIG     111  /* array dimensions exceed internal limit */
#define READONLY_FILE     112  /* Cannot write to readonly file */
#define MEMORY_ALLOCATION 113  /* Could not allocate memory */
#define BAD_FILEPTR       114  /* invalid fitsfile pointer */
#define NULL_INPUT_PTR    115  /* NULL input pointer to routine */
#define SEEK_ERROR        116  /* error seeking position in file */

#define BAD_URL_PREFIX    121  /* invalid URL prefix on file name */
#define TOO_MANY_DRIVERS  122  /* tried to register too many IO drivers */
#define DRIVER_INIT_FAILED 123  /* driver initialization failed */
#define NO_MATCHING_DRIVER 124  /* matching driver is not registered */
#define URL_PARSE_ERROR    125  /* failed to parse input file URL */

#define HEADER_NOT_EMPTY  201  /* header already contains keywords */
#define KEY_NO_EXIST      202  /* keyword not found in header */
#define KEY_OUT_BOUNDS    203  /* keyword record number is out of bounds */
#define VALUE_UNDEFINED   204  /* keyword value field is blank */
#define NO_QUOTE          205  /* string is missing the closing quote */
#define BAD_KEYCHAR       207  /* illegal character in keyword name or card */
#define BAD_ORDER         208  /* required keywords out of order */
#define NOT_POS_INT       209  /* keyword value is not a positive integer */
#define NO_END            210  /* couldn't find END keyword */
#define BAD_BITPIX        211  /* illegal BITPIX keyword value*/
#define BAD_NAXIS         212  /* illegal NAXIS keyword value */
#define BAD_NAXES         213  /* illegal NAXISn keyword value */
#define BAD_PCOUNT        214  /* illegal PCOUNT keyword value */
#define BAD_GCOUNT        215  /* illegal GCOUNT keyword value */
#define BAD_TFIELDS       216  /* illegal TFIELDS keyword value */
#define NEG_WIDTH         217  /* negative table row size */
#define NEG_ROWS          218  /* negative number of rows in table */
#define COL_NOT_FOUND     219  /* column with this name not found in table */
#define BAD_SIMPLE        220  /* illegal value of SIMPLE keyword  */
#define NO_SIMPLE         221  /* Primary array doesn't start with SIMPLE */
#define NO_BITPIX         222  /* Second keyword not BITPIX */
#define NO_NAXIS          223  /* Third keyword not NAXIS */
#define NO_NAXES          224  /* Couldn't find all the NAXISn keywords */
#define NO_XTENSION       225  /* HDU doesn't start with XTENSION keyword */
#define NOT_ATABLE        226  /* the CHDU is not an ASCII table extension */
#define NOT_BTABLE        227  /* the CHDU is not a binary table extension */
#define NO_PCOUNT         228  /* couldn't find PCOUNT keyword */
#define NO_GCOUNT         229  /* couldn't find GCOUNT keyword */
#define NO_TFIELDS        230  /* couldn't find TFIELDS keyword */
#define NO_TBCOL          231  /* couldn't find TBCOLn keyword */
#define NO_TFORM          232  /* couldn't find TFORMn keyword */
#define NOT_IMAGE         233  /* the CHDU is not an IMAGE extension */
#define BAD_TBCOL         234  /* TBCOLn keyword value < 0 or > rowlength */
#define NOT_TABLE         235  /* the CHDU is not a table */
#define COL_TOO_WIDE      236  /* column is too wide to fit in table */
#define COL_NOT_UNIQUE    237  /* more than 1 column name matches template */
#define BAD_ROW_WIDTH     241  /* sum of column widths not = NAXIS1 */
#define UNKNOWN_EXT       251  /* unrecognizable FITS extension type */
#define UNKNOWN_REC       252  /* unrecognizable FITS record */
#define END_JUNK          253  /* END keyword is not blank */
#define BAD_HEADER_FILL   254  /* Header fill area not blank */
#define BAD_DATA_FILL     255  /* Data fill area not blank or zero */
#define BAD_TFORM         261  /* illegal TFORM format code */
#define BAD_TFORM_DTYPE   262  /* unrecognizable TFORM datatype code */
#define BAD_TDIM          263  /* illegal TDIMn keyword value */
 
#define BAD_HDU_NUM       301  /* HDU number < 1 or > MAXHDU */
#define BAD_COL_NUM       302  /* column number < 1 or > tfields */
#define NEG_FILE_POS      304  /* tried to move before beginning of file  */
#define NEG_BYTES         306  /* tried to read or write negative bytes */
#define BAD_ROW_NUM       307  /* illegal starting row number in table */
#define BAD_ELEM_NUM      308  /* illegal starting element number in vector */
#define NOT_ASCII_COL     309  /* this is not an ASCII string column */
#define NOT_LOGICAL_COL   310  /* this is not a logical datatype column */
#define BAD_ATABLE_FORMAT 311  /* ASCII table column has wrong format */
#define BAD_BTABLE_FORMAT 312  /* Binary table column has wrong format */
#define NO_NULL           314  /* null value has not been defined */
#define NOT_VARI_LEN      317  /* this is not a variable length column */
#define BAD_DIMEN         320  /* illegal number of dimensions in array */
#define BAD_PIX_NUM       321  /* first pixel number greater than last pixel */
#define ZERO_SCALE        322  /* illegal BSCALE or TSCALn keyword = 0 */
#define NEG_AXIS          323  /* illegal axis length < 1 */
 
#define NOT_GROUP_TABLE         340
#define HDU_ALREADY_MEMBER      341
#define MEMBER_NOT_FOUND        342
#define GROUP_NOT_FOUND         343
#define BAD_GROUP_ID            344
#define TOO_MANY_HDUS_TRACKED   345
#define HDU_ALREADY_TRACKED     346
#define BAD_OPTION              347
#define IDENTICAL_POINTERS      348

#define BAD_I2C           401  /* bad int to formatted string conversion */
#define BAD_F2C           402  /* bad float to formatted string conversion */
#define BAD_INTKEY        403  /* can't interprete keyword value as integer */
#define BAD_LOGICALKEY    404  /* can't interprete keyword value as logical */
#define BAD_FLOATKEY      405  /* can't interprete keyword value as float */
#define BAD_DOUBLEKEY     406  /* can't interprete keyword value as double */
#define BAD_C2I           407  /* bad formatted string to int conversion */
#define BAD_C2F           408  /* bad formatted string to float conversion */
#define BAD_C2D           409  /* bad formatted string to double conversion */
#define BAD_DATATYPE      410  /* bad keyword datatype code */
#define BAD_DECIM         411  /* bad number of decimal places specified */
#define NUM_OVERFLOW      412  /* overflow during datatype conversion */
#define BAD_DATE          420  /* error in date or time conversion */

#define PARSE_SYNTAX_ERR  431  /* syntax error in parser expression */
#define PARSE_BAD_TYPE    432  /* expression did not evaluate to desired type */
#define PARSE_LRG_VECTOR  433  /* vector result too large to return in array */
#define PARSE_NO_OUTPUT   434  /* data parser failed not sent an out column */
#define PARSE_BAD_COL     435  /* bad data encounter while parsing column */
#define PARSE_BAD_OUTPUT  436  /* Output file not of proper type          */

#define ANGLE_TOO_BIG     501  /* celestial angle too large for projection */
#define BAD_WCS_VAL       502  /* bad celestial coordinate or pixel value */
#define WCS_ERROR         503  /* error in celestial coordinate calculation */
#define BAD_WCS_PROJ      504  /* unsupported type of celestial projection */
#define NO_WCS_KEY        505  /* celestial coordinate keywords not found */

/*------- following error codes are used in the grparser.c file -----------*/
#define	NGP_ERRBASE		(360)			/* base chosen so not to interfere with CFITSIO */
#define	NGP_OK			(0)
#define	NGP_NO_MEMORY		(NGP_ERRBASE + 0)	/* malloc failed */
#define	NGP_READ_ERR		(NGP_ERRBASE + 1)	/* read error from file */
#define	NGP_NUL_PTR		(NGP_ERRBASE + 2)	/* null pointer passed as argument */
#define	NGP_EMPTY_CURLINE	(NGP_ERRBASE + 3)	/* line read seems to be empty */
#define	NGP_UNREAD_QUEUE_FULL	(NGP_ERRBASE + 4)	/* cannot unread more then 1 line (or single line twice) */
#define	NGP_INC_NESTING		(NGP_ERRBASE + 5)	/* too deep include file nesting (inf. loop ?) */
#define	NGP_ERR_FOPEN		(NGP_ERRBASE + 6)	/* fopen() failed, cannot open file */
#define	NGP_EOF			(NGP_ERRBASE + 7)	/* end of file encountered */
#define	NGP_BAD_ARG		(NGP_ERRBASE + 8)	/* bad arguments passed */
#define	NGP_TOKEN_NOT_EXPECT	(NGP_ERRBASE + 9)	/* token not expected here */

/*------------------- a few definitions to read FITS parameters ------------*/

#define FBSIZE  2880L   /* size (in bytes) of one FITS block */
#define FITSTOF(k, def) ((point = fitsnfind(buf, k, n))? \
                                 atof(strncpy(st, &point[10], 70)) \
                                :(def))
#define FITSTOI(k, def) ((point = fitsnfind(buf, k, n))? \
                                 atoi(strncpy(st, &point[10], 70)) \
                                :(def))
#define FITSTOS(k, str, def) { point = fitsnfind(buf, k, n); \
                               if (point != NULL) { \
                                  for (i=0,point+=11; (*point)!='\'' && i < 69;) \
                                     (str)[i++] = *(point++); \
                                  (str)[i] = '\0'; \
                                } else\
                                strcpy(str, def); \
                             }
#define QFREAD(ptr, size, file, fname) \
                if (fread(ptr, (size_t)(size), (size_t)1, file)!=1) \
                  error(EXIT_FAILURE, "*Error* while reading ", fname)

#define QFWRITE(ptr, size, file, fname) \
                if (fwrite(ptr, (size_t)(size), (size_t)1, file)!=1) \
                  error(EXIT_FAILURE, "*Error* while writing ", fname)

#define QFSEEK(file, offset, pos, fname) \
                if (fseek(file, (offset), pos)) \
                  error(EXIT_FAILURE,"*Error*: file positioning failed in ", \
                        fname)
#define QFTELL(pos, file, fname) \
                if ((pos=ftell(file))==-1) \
                  error(EXIT_FAILURE,"*Error*: file position unknown in ", \
                        fname)
/* int     t_size[] = {1, 2, 4, 4, 8}; */
typedef enum {H_INT, H_FLOAT, H_EXPO, H_BOOL, H_STRING, H_COMMENT,
	      H_KEY}  h_type;         /* type of FITS-header data */


enum type_data {T_BYTE, T_SHORT, T_INT, T_FLOAT, T_DOUBLE,
		T_COMPLEX_F, T_COMPLEX_D, UN_KNOWN};

typedef unsigned char byte;
typedef unsigned long u_long;

/*--------------------------- FITS BitPix coding ----------------------------*/
#define BP_BYTE         8
#define BP_SHORT        16
#define BP_INT          32
#define BP_FLOAT        (-32)
#define BP_DOUBLE       (-64)

#define MAX_NBR_AXIS 	3 /* Maximum number of axes */
#define FLOAT_EPSILON 5.96047e-08

#define DBUFFSIZE 28800 /* size of data buffer in bytes */

#define NIOBUF 25        /* number of IO buffers to create */
#define IOBUFLEN 2880    /* size in bytes of each IO buffer */
#define MINDIRECT 8640   /* minimum size for direct reads and writes */
                         /* MINDIRECT must have a value >= 8640 */

#define NATIVE             0 /* a generic machine that uses IEEE formats */
#define ULTRIX             1
#define ALPHA_OSF          2
#define VAXVMS             3
#define ALPHAVMS           4
#define IBMPC              5
#define CRAY               6

#define GFLOAT             1
#define IEEEFLOAT          2

/* the following are used to determine what type machine we are running on */

/* the following block determines the size of longs on SGI IRIX machines */
#if defined(_MIPS_SZLONG)
#  if _MIPS_SZLONG == 32
#    define LONGSIZE 32
#  elif _MIPS_SZLONG == 64
#    define LONGSIZE 64
#  else
#    error "can't handle long size given by _MIPS_SZLONG"
#  endif
#endif

#if defined(vax) && defined(VMS)
 
#define MACHINE VAXVMS
#define BYTESWAPPED 1
 
#elif defined(__alpha) && defined(__VMS) &&  (__D_FLOAT == TRUE)

/* this float option is the same as for VAX/VMS machines. */
#define MACHINE VAXVMS
#define BYTESWAPPED 1
 
#elif defined(__alpha) && defined(__VMS) &&  (__G_FLOAT == TRUE)
 
/*  G_FLOAT is the default for ALPHA VMS systems */
#define MACHINE ALPHAVMS
#define BYTESWAPPED 1
#define FLOATTYPE GFLOAT
 
#elif defined(__alpha) && defined(__VMS) &&  (__IEEE_FLOAT == TRUE)
 
#define MACHINE ALPHAVMS
#define BYTESWAPPED 1
#define FLOATTYPE IEEEFLOAT

#elif defined(__alpha) && defined(__unix__)
 
#define MACHINE ALPHA_OSF
#define BYTESWAPPED 1
#define LONGSIZE 64

#elif defined(ultrix) && defined(unix)
 
#define MACHINE ULTRIX
#define BYTESWAPPED 1
 
#elif defined(__i386) || defined(__i386__) || defined(__i486__) || defined(__i586__) 

/*  IBM PC */
#define MACHINE IBMPC
#define BYTESWAPPED 1

#elif defined(_MSC_VER) || defined(__BORLANDC__) || defined(__TURBOC__)

/*  IBM PC running DOS or Windows */
#define MACHINE IBMPC       
#define BYTESWAPPED 1

#elif defined(_NI_mswin_)

/*  LabWindows/CVI with Windows 3.x, 95, or NT  */
#define MACHINE IBMPC       
#define BYTESWAPPED 1

#elif  defined(__EMX__)

/*  IBM PC running OS/2 */
#define MACHINE IBMPC
#define BYTESWAPPED 1

#else

/*  assume machine uses the same IEEE formats as used in FITS files */
#define MACHINE NATIVE
#define BYTESWAPPED 0
#endif


/*  assume longs are 4 bytes long, unless previously set otherwise */
#ifndef LONGSIZE
#define LONGSIZE 32
#endif
  
#define IGNORE_EOF 1
#define REPORT_EOF 0
#define DATA_UNDEFINED -1
#define NULL_UNDEFINED 1234554321
#define ASCII_NULL_UNDEFINED 1   /* indicate no defined null value */
 
#define maxvalue(A,B) ((A) > (B) ? (A) : (B))
#define minvalue(A,B) ((A) < (B) ? (A) : (B))
 
#define FSTRCMP(a,b)     ((a)[0]<(b)[0]? -1:(a)[0]>(b)[0]?1:strcmp((a),(b)))
#define FSTRNCMP(a,b,n)  ((a)[0]<(b)[0]?-1:(a)[0]>(b)[0]?1:strncmp((a),(b),(n)))

#if MACHINE == ALPHAVMS
 
#define FNANMASK   0x7F00 /* mask bits 1 - 7; all set on NaNs */
                                     /* all 0 on underflow  or 0. */
 
#define DNANMASK   0x7FE0 /* mask bits 1 - 10; all set on NaNs */
                                     /* all 0 on underflow  or 0. */
 
#else
 
#define FNANMASK   0x7F80 /* mask bits 1 - 8; all set on NaNs */
                                     /* all 0 on underflow  or 0. */
 
#define DNANMASK   0x7FF0 /* mask bits 1 - 11; all set on NaNs */
                                     /* all 0 on underflow  or 0. */
 
#endif

#if MACHINE == CRAY
    /*
      Cray machines:   the large negative integer corresponds
      to the 3 most sig digits set to 1.   If these
      3 bits are set in a floating point number (64 bits), then it represents
      a reserved value (i.e., a NaN)
    */
#define fnan(L) ( (L) >= 0xE000000000000000 ? 1 : 0)
#else
    /* these functions work for both big and little endian machines */
    /* that use the IEEE floating point format for internal numbers */
 
   /* These functions tests whether the float value is a reserved IEEE     */
   /* value such as a Not-a-Number (NaN), or underflow, overflow, or       */
   /* infinity.   The functions returns 1 if the value is a NaN, overflow  */
   /* or infinity; it returns 2 if the value is an denormalized underflow  */
   /* value; otherwise it returns 0. fnan tests floats, dnan tests doubles */
 
#define fnan(L) \
      ( (L & FNANMASK) == FNANMASK ?  1 : (L & FNANMASK) == 0 ? 2 : 0)
 
#define dnan(L) \
      ( (L & DNANMASK) == DNANMASK ?  1 : (L & DNANMASK) == 0 ? 2 : 0)
 
#endif

#define DUCHAR_MAX  255.49 /* max double value that fits in an unsigned char */
#define DUCHAR_MIN -0.49   /* min double value that fits in an unsigned char */
#define DUSHRT_MAX  65535.49 /* max double value that fits in a unsigned short*/
#define DUSHRT_MIN -0.49   /* min double value that fits in an unsigned short */
#define DSHRT_MAX  32767.49 /* max double value that fits in a short */
#define DSHRT_MIN -32768.49 /* min double value that fits in a short */
#define DULONG_MAX 4294967295.49 /* max double that fits in a unsigned long */
#define DULONG_MIN -0.49   /* min double value that fits in an unsigned long */
#define DLONG_MAX  2147483647.49 /* max double value that fits in a long */
#define DLONG_MIN -2147483648.49 /* min double value that fits in a long */
#define DUINT_MAX 4294967295.49 /* max dbl that fits in a unsigned 4-byte int */
#define DUINT_MIN -0.49   /* min dbl that fits in an unsigned 4-byte int */
#define DINT_MAX  2147483647.49 /* max double value that fits in a 4-byte int */
#define DINT_MIN -2147483648.49 /* min double value that fits in a 4-byte int */

#ifndef UINT32_MAX
#define UINT32_MAX 4294967295U /* max unsigned 32-bit integer */
#endif
#ifndef INT32_MAX
#define INT32_MAX  2147483647 /* max 32-bit integer */
#endif
#ifndef INT32_MIN
#define INT32_MIN -2147483647 /* min 32-bit integer */
#endif

/* ==================== SHARED MEMORY DRIVER SECTION ======================= */

#ifdef HAVE_SHMEM_SERVICES
#include "drvrsmem.h"
#endif

/* ==================== END OF SHARED MEMORY DRIVER SECTION ================ */

#if defined(vms) || defined(__vms) || defined(__WIN32__) || defined(macintosh)

/* ================================================================== */
/* A hack for nonunix machines, which lack strcasecmp and strncasecmp */
/* ================================================================== */

int strcasecmp (char *s1, char *s2       );
int strncasecmp(char *s1, char *s2, int n);

#endif

/*----------------------------- Fits image parameters ----------------------------*/

// typedef struct
class fitsstruct {
  void fitsinit() {
    file=NULL;	
    fitsheadsize= 2880;
    bitpix = 0;
    bytepix = 0;
    width = 0;
    height = 0;
    npix = 0;
    bscale = 1.;
    bzero = 0.;
    crpixx = 0.;
    crpixy = 0.;
    crvalx = 0.;
    crvaly = 0.;
    cdeltx= 0.;
    cdelty = 0.;
    ngamma=0.;
    pixscale=0.;
    nlevels = 0;
    pixmin  = 0.;
    pixmax  = 0.;
    epoch = 0.;
    crotax = 0.;
    crotay = 0.;
    fitshead = NULL;
    origin = "";
    strcpy(ctypex, "");
    strcpy(ctypey, "");
    strcpy(rident, "");
    
    /* HISTORY & COMMENT fields */
    history = NULL;
    hist_size = 0;
    comment = NULL;
    com_size = 0;
    
    naxis=0;
    for (int i=0; i < MAX_NBR_AXIS;i++) {
      TabAxis[i]=0;
      TabStep[i]=0.;
      TabRef[i]=0.;
      TabValRef[i]=0.;
    }
    filename = NULL;
    origin = NULL;
    fitshead = NULL;
  }
 public:
  
  fitsstruct () { 
    fitsinit();
  }
  /*void hd_fltarray(fltarray &Mat, char *History=NULL) {
    char *creafitsheader();
    
    fitshead = creafitsheader();
    bitpix = -32;
    width =  Mat.nx();
    height = Mat.ny();
    naxis = Mat.naxis();
    npix = Mat.n_elem();
    for (int i=0; i < naxis; i++) TabAxis[i] = Mat.axis(i+1);
    if (History != NULL) origin = History;
    }*/
  
  ~fitsstruct() { 
    if (filename != NULL) free (filename);
    // if (fitshead != NULL) free ((char *) fitshead);
    if (history != NULL)  free (history);
    if (comment != NULL)  free (comment);
    fitsinit();
  }
  
  char		*filename;		/* pointer to the image filename */
  char          *origin;                /* pointer to the origin */
  char		ident[512];		/* field identifier (read from FITS)*/
  char		rident[512];	        /* field identifier (relative) */
  FILE		*file;			/* pointer the image file structure */
  char		*fitshead;		/* pointer to the FITS header */
  int		fitsheadsize;		/* FITS header size */
  /* ---- main image parameters */
  int		bitpix, bytepix;	/* nb of bits and bytes per pixel */
  int		width, height;		/* x,y size of the field */
  int		npix;			/* total number of pixels */
  double	bscale, bzero;		/* FITS scale and offset */
  double	ngamma;			/* normalized photo gamma */
  int		nlevels;		/* nb of quantification levels */
  float		pixmin, pixmax;		/* min and max values in frame */
  /* ---- basic astrometric parameters */
  double	epoch;			/* epoch for coordinates */
  double	pixscale;		/* pixel size in arcsec.pix-1 */
					/* */
  /* ---- astrometric parameters */
  double	crpixx,crpixy;		/* FITS CRPIXn */
  double	crvalx,crvaly;		/* FITS CRVALn */
  double	cdeltx,cdelty;		/* FITS CDELTn */
  double	crotax,crotay;		/* FITS CROTAn */
  char          ctypex[256];            /* FITS CTYPE1 */
  char          ctypey[256];            /* FITS CTYPE2 */
  char          CoordType[256];         
  
  /* ---- HISTORY & COMMENT parameters --- */
  char *history;
  int hist_size;
  char *comment;
  int com_size;
  
  /* ---- for non image use */
  int naxis;
  int TabAxis[MAX_NBR_AXIS];
  double TabStep[MAX_NBR_AXIS];
  double TabRef[MAX_NBR_AXIS];
  double TabValRef[MAX_NBR_AXIS];
};


/* FITS read-write utilities. */

enum Bool {False = 0,True = 1};
int DEBUH = 0;
using namespace std;

void error(int num, char *msg1, char *msg2) {
  fprintf(stderr, "Error: %s%s\n",msg1,msg2);
  exit(num);
}

/*
 * Swap bytes for doubles, longs and shorts (for DEC machines or PC for inst.).
 */
void swapbytes(void *ptr, int nb, int n) {
  char	*cp;
  int j;
  
  cp = (char *)ptr;
  
  if (nb&4) {
    for (j=0; j<n; j++) {
      cp[0] ^= (cp[3]^=(cp[0]^=cp[3]));
      cp[1] ^= (cp[2]^=(cp[1]^=cp[2]));
      cp +=4;
    }
    return;
  }
  if (nb&2) {
    for (j=0; j<n; j++) {
      cp[0] ^= (cp[1]^=(cp[0]^=cp[1]));
      cp += 2;
    }
    return;
  }
  if (nb&1)
    return;
  
  if (nb&8) {
    for (j=0; j<n; j++) {
      cp[0] ^= (cp[7]^=(cp[0]^=cp[7]));
      cp[1] ^= (cp[6]^=(cp[1]^=cp[6]));
      cp[2] ^= (cp[5]^=(cp[2]^=cp[5]));
      cp[3] ^= (cp[4]^=(cp[3]^=cp[4]));
      cp += 8;
    }
    return;
  }
  
  error(EXIT_FAILURE, "*Internal Error*: Unknown size in ", "swapbytes()");
  
  return;
}

/**
 * : ( 0 1 -> 1 0 )
 */
void ffswap2 ( short *values, long nvalues ) {
   register char *cvalues;
   register long ii;
   
   union u_tag {
      char cvals[2];   /* equivalence an array of 4 bytes with */
      short sval;      /* a short */
   } u;
   
   cvalues = (char *) values;      /* copy the initial pointer value */
   
   for (ii = 0; ii < nvalues;) {
      u.sval = values[ii++];  /* copy next short to temporary buffer */
      
      *cvalues++ = u.cvals[1]; /* copy the 2 bytes to output in turn */
      *cvalues++ = u.cvals[0];
   }
   return;
}

/**
 * : ( 0 1 2 3 -> 3 2 1 0 )
 */
void ffswap4 ( INT32BIT *values, long nvalues ) {
   register char *cvalues;
   register long ii;
   
   union u_tag {
      char cvals[4];      /* equivalence an array of 4 bytes with */
      INT32BIT ival;      /* a float */
   } u;
   
   cvalues = (char *) values;   /* copy the initial pointer value */
   
   for (ii = 0; ii < nvalues;) {
      u.ival = values[ii++];  /* copy next float to buffer */
      
      *cvalues++ = u.cvals[3]; /* copy the 4 bytes in turn */
      *cvalues++ = u.cvals[2];
      *cvalues++ = u.cvals[1];
      *cvalues++ = u.cvals[0]; 
   }
   return;
}

/**
 * : ( 01234567  -> 76543210 )
 */
void ffswap8 (double *values, long nvalues) {
   register char *cvalues;
   register long ii;
   register char temp;
   
   cvalues = (char *) values;      /* copy the pointer value */
   
   for (ii = 0; ii < nvalues*8; ii += 8) {
      temp = cvalues[ii];
      cvalues[ii] = cvalues[ii+7];
      cvalues[ii+7] = temp;
      
      temp = cvalues[ii+1];
      cvalues[ii+1] = cvalues[ii+6];
      cvalues[ii+6] = temp;
      
      temp = cvalues[ii+2];
      cvalues[ii+2] = cvalues[ii+5];
      cvalues[ii+5] = temp;
      
      temp = cvalues[ii+3];
      cvalues[ii+3] = cvalues[ii+4];
      cvalues[ii+4] = temp;
   }
   return;
}

/**
 * Complete the name of the file with the
 * suitable extension if needed.
 * If NameStep==NULL returns NULL.
 * @param NameStep the input file name.
 * @return the complete file name.
 */
char *fitsname(char * NameStep) {
   char Sreturn[512];
   char Name[512];
   char *ValRet;
   
   if ((!NameStep)) return NULL;
   strcpy(Name, NameStep);
   if (strstr(Name, ".fit") != NULL)  strcpy(Sreturn, Name);
   else sprintf(Sreturn, "%s.%s", Name, "fits");
   ValRet = strdup(Sreturn);
   return (ValRet);
}

Bool std_inout(char * /*filename*/) {
   return False;
}

FILE *fits_file_des_in(char *fname) {
   FILE *fp;
   char *realname= fitsname(fname);
   if (std_inout(fname) == True)  fp = stdin;
   else  {
      fp = fopen(fitsname(fname), "rb");
      if (!fp) {
	 cerr << "Error: Unable to open file: " <<  realname  << endl;
	 exit(-1);
      }
   }
   return fp;
}

/**
 * Opens the file from the specified filename.
 * If (filename = NULL) or (filename ="-") opens the standard input.
 * @param fname	the file name.
 * @return the file descriptor.
 */
FILE *fits_file_des_out(char *fname) {
   FILE *fp;
   
   if (!fname || strstr(fname,"-")) {
      return stdout;
   }
   
   if (std_inout(fname) == True) fp = stdout;
   else  {
      fp = fopen(fitsname(fname), "wb");
      if (!fp) {
	 cerr << "Error: Unable to create file: " <<  fname  << endl;
	 exit(-1);
      }
   }
   return fp;
} 

/*
 * Search for a FITS keyword in a FITS header. 
 * @return the first line containing the keyword: keyword
 */
int fitsfind(char *fitsbuf, char *keyword) {
   char	*ptr;
   int	i=0, len;
   
   len = strlen(keyword);
   
   ptr= (char *) &(fitsbuf[80*i]);
   // RG printf(" %80s\n", ptr);
   for (i=0; strncmp(ptr, "END     ", 8); i++) {
      if (!strncmp(ptr, keyword, len)) {
	 // printf(" we got it %d\n",i);
	 return i;
      }
      ptr= (char *) &(fitsbuf[80*(i+1)]);
   }
   if (strncmp(keyword, "END     ", 8)) {
      return -1;
   } else {
      return i;
   }
}

char *creafitsheader() {
   char *Buf;
   int i,k;
   static char	bufmodel[4][80] = {
      "SIMPLE  =                    T / THIS IS REGULAR FITS",
      "BITPIX  =                    0 /",
      "NAXIS   =                    0 /",
      "END"};
   
   if (!(Buf = (char *) malloc (FBSIZE*sizeof(char))))
      error(EXIT_FAILURE, "*Error*: Not enough memory in ", "creafitsheader()");
  
   for (k = 0; k < 4; k++) {
      for (i = 0; i < (int) strlen (bufmodel[k]); i++) Buf[k*80+i] = bufmodel[k][i];
      for (i = strlen (bufmodel[k]); i < 80; i++) Buf[k*80+i] = ' ';
      
   }
   for (k = 4; k < 36; k++) for (i = 0; i < 80; i++) Buf[k*80+i] = ' ';
   return (Buf);
}

/* 
 * PURPOSE: Add a comment in the field 'history' or 'comment' of the structure 
 *    FITSSTRUCT. It begins at new line of 72 characters.
 *
 * INPUT:  comment --- STRING, value of HISTORY or COMMENT field;
 *         type_com --- STRING, "HISTORY" or "COMMENT" to specify in which field the comment
 *                      must be appended.
 * OUTPUT: pfitsbuf --- FITSSTRUCT *, address of the structure FITSSTRUCT holding the FITS Header;
 * RETURN: INT, the length of the added comment.
 *
 *  16-MAR-1998 
 */
int fitsaddhist_com(fitsstruct *pfitsbuf, char *comment, char *type_com) {
  int add_len = 0;  /* Length of the comment or history to be added */
  int nb_chars = 0; /* number of characters stored in out */
  int out_nblines = 0;  /* number of lines of the output */
  int add_nblines = 0;  /* number of lines to be added */
  char *out_ptr=NULL; /* pointer to the output */
  int out_size=0;   /* the allocated size, not to be confused with the length! */
  int out_len ;  /* the actual size */
  int status = 0;
  char *Newptr;
  int i;
  
  /* Length of the comment to be added */
  if (comment != NULL) add_len = strlen(comment); 
  else return(0); /* it is finished nothing to add !!! */
  
  /*--- TYPE of the comment ---*/
  if ( strcmp(type_com,"HISTORY") == 0 ) {
    out_ptr        = pfitsbuf->history;
    out_size = pfitsbuf->hist_size; 
    status = 1;
  }
  
  if ( strcmp(type_com,"COMMENT") == 0 ) {
    out_ptr        = pfitsbuf->comment;
    out_size = pfitsbuf->com_size; 
    status = 1;
  }
  
  if ( status == 0 ) {
    printf(" ERROR in fitsaddhist_com: the type of the comment is  %s!\n\n",
	   type_com);
    exit(-1);
  }
  
  /* find out the length of the field */
  if (out_ptr == NULL) out_len = 0; else out_len = strlen(out_ptr);
  
  //  nb_chars = out_len + add_len;
  out_nblines = out_len / 72;
  if( (out_len % 72) > 0) out_nblines ++;
  nb_chars = (out_len % 72) ;
  add_nblines = add_len / 72;
  if( (add_len % 72) > 0) add_nblines ++;
  out_nblines += add_nblines;
  
  /* Memory reallocation if there is not enough memory for adding the comment */
  if (out_nblines*72 +1  > out_size) {
    //  Newptr = new char [out_nblines*72 +1  ];
    Newptr = (char*) malloc( (out_nblines*72 +1  ) *sizeof(char));
    if (Newptr == NULL) {
      cerr << "Error: unable to allocate memory : "  << endl;
      exit(-1);
    }
    memset(Newptr,'\0', (out_nblines*72 +1));
    strncpy(Newptr, out_ptr, out_size);
    // Newptr[out_size] = '\0';
    free ((char *) out_ptr);
    out_ptr = Newptr;
    out_size = out_nblines*72 +1;
    if (DEBUH) cout << "Memory reallocation = " << out_size << endl;
    if (DEBUH) cout << " out_ptr = " << out_ptr <<"X" << endl;
  }
  if (DEBUH) cout << " nb_chars="  << nb_chars << endl;
  if (nb_chars > 0) for(i = nb_chars; i < 72;i++) strcat(out_ptr, " ");
  /* Now add the comment */
  strcat(out_ptr, comment);
  
  if (DEBUH) cout << " out_ptr = " << out_ptr  << endl;
  
  /* update of the input structure */
  if ( strcmp(type_com,"COMMENT") == 0 ) {
    pfitsbuf->comment = out_ptr ;
    pfitsbuf->com_size = out_size; 
  } else {
    pfitsbuf->history = out_ptr ;
    pfitsbuf->hist_size = out_size; 
  }
  return add_len;
  
}

/* 
 * PURPOSE: add in the field 'fitshead' the value contained in the fields 'history' & 'comment'.
 *          The header size referred by 'pfitsbuf->fitsheadsize' is updated.
 *
 * INPUTS: pfitsbuf --- POINTER to FITSSTRUCT, the fits header to update
 *
 *
 * RETURN: INT, the size of 'fitshead'.
 *
 */ 
int fitswritehist(fitsstruct *pfitsbuf) {
  char    *outptr, *ptr, *curptr;
  int     len, lenhist, lencom,size,i;
  
  size = pfitsbuf->fitsheadsize;
  outptr = (char *) calloc(size+1 ,sizeof(char)); // v3.4
  
  /* copy all the fields, except HISTORY & COMMENT fields, of the fits header in the new string */
  ptr = pfitsbuf->fitshead;
  curptr = outptr;
  while (strncmp(ptr,"END",3) != 0) {
    if ((strncmp(ptr,"HISTORY",7) != 0) && (strncmp(ptr,"COMMENT",7) != 0)){
      strncpy(curptr,ptr,80);
      curptr +=80;
    }
    ptr +=80;
  }
  *curptr='\0';
  
  /* Is there enough memory for adding 'history' & 'comment' fields
   * of the FITSSTRUCT structure? */
  lenhist = (pfitsbuf->history != NULL?strlen(pfitsbuf->history):0);
  lencom = (pfitsbuf->comment != NULL? strlen(pfitsbuf->comment) :0);
  len =  lenhist + ((lenhist / 72)+1)*8 + lencom + ((lencom / 72)+1)*8 + strlen(outptr) + 80*sizeof(char);
  if (len >= FBSIZE) {
    size = FBSIZE*(len/FBSIZE + 1);
    outptr = (char *) realloc( (void *) outptr, (size_t) sizeof(char)*(size+1) );
    curptr = outptr + strlen(outptr);
  }
    
  /* Add in 'fitshead' field the 'history' & 'comment' fields of the FITSSTRUCT structure */
  len = (pfitsbuf->history != NULL?strlen(pfitsbuf->history):0);
  ptr = pfitsbuf->history;
  i=0;
  while(i<len){
    if( ptr + 72 > pfitsbuf->history + pfitsbuf->hist_size ) 
      printf(" warning fitswritehist hist_size=%d \n",pfitsbuf->hist_size );
    sprintf(curptr,"%-8.8s%-72.72s","HISTORY ",ptr);
    i += 72;
    ptr += 72;
    curptr += 80;
  }
  
  len = (pfitsbuf->comment != NULL? strlen(pfitsbuf->comment) :0);
  ptr = pfitsbuf->comment;
  i=0;
  // printf(" pfitsbuf->com_size=%d len=%d\n", pfitsbuf->com_size, len);
  while(i<len){
    // printf(" strlen=%d \n", strlen(ptr));
    if( ptr +72 > pfitsbuf->comment + pfitsbuf->com_size ) 
      printf(" warning fitswritehist com_size=%d \n",pfitsbuf->com_size );
    sprintf(curptr,"%-8.8s%-72.72s","COMMENT ",ptr);
    // left centered, 8 is the minimum number of caracters
    //  .8 is the maximum number of carac.
    
    i += 72;
    ptr += 72;
    curptr += 80;
  }
  
  /* Add the 'END' keyword */
  sprintf(curptr,"%-80s","END");
  len = (size - strlen(outptr))/80;
  
  for (i=0, curptr += 80;i<len;i++,curptr += 80) 
    sprintf(curptr,"%-80s","  ");
  
  /* Update the  header and its size */
  pfitsbuf->fitshead = outptr;
  pfitsbuf->fitsheadsize = size;
  
  return size; 
}

/* 
 * PURPOSE: If the keyword 'keyword' doesn't exist, a fits line containing
 * this keyword is created & added to the STRING referred by 'add_fitsbuf'.
 * The header size referred by pfitsheadsize
 *
 * INPUTS: keyword --- STRING, the name of the keyword to update or to add
 *         comment --- STRING, value to put in the field 'keyword'
 *
 * OUTPUTS: add_fitsbuf   --- STRING POINTER, the address of the fits header string to update
 *          pfitsheadsize --- POINTER, address of the header size
 *
 * RETURN: INT, the line number in the fits header where the keyword has been added.
 *
 * CAUTION: doesn't work for creating multiple fits lines field like HISTORY or COMMENT fields.
 */
int fitsadd(char **add_fitsbuf, char *keyword, char *comment, int *pfitsheadsize) {
  char    *key_ptr;
  int     headpos, nkey;

  if ((headpos = fitsfind(*add_fitsbuf, keyword)) < 0) {
    headpos = fitsfind(*add_fitsbuf, "END     ");
    nkey = *pfitsheadsize / 80;
    if (nkey - headpos < 4) {
      *pfitsheadsize += FBSIZE;
      *add_fitsbuf = (char *) realloc((char *) *add_fitsbuf, (size_t) *pfitsheadsize );
      memset(*add_fitsbuf+*pfitsheadsize-FBSIZE,'\0', FBSIZE);  // v3.4
      
    }
    
    // Go to the 'END' line and copy it to the next line
    key_ptr = *add_fitsbuf + 80*headpos;
    
    sprintf(key_ptr, "%-8.8s=                      / %-47.47s",
	    keyword, comment?comment:" ");
    sprintf(key_ptr + 80, "%-80s","END");
    
  }
  
  return headpos;
}

/* 
 * PURPOSE: Update the value (enclosed in 'ptr') of the keyword 'keyword' in 
 * the STRING referred by 'fitsbuf'.
 * If the keyword 'keyword' doesn't exist, a fits line containing this keyword
 * is created & added to the STRING referred by 'fitsbuf'.
 * The header size referred by pfitsheadsize
 *
 * INPUTS: keyword --- STRING, the name of the keyword to update or to add
 *         ptr     --- POINTER, address of the value to put in the field 'keyword'
 *         type    --- H_TYPE, type of the value referred by 'ptr'
 *         t_type  --- TYPE_DATA, secondary type of the value referred by 'ptr'
 *
 * OUTPUTS: fitsbuf --- STRING, the fits header to update
 *          pfitsheadsize --- POINTER, address of the header size
 *
 * RETURN: INT, SUCCESS if no problems, FAILURE if an error occures.
 *
 * CAUTION: doesn't work for updating or creating multiple fits lines
 * field like HISTORY or COMMENT fields.
 */ 
int fitswrite(char *fitsbuf, char *keyword, void *ptr, h_type type, type_data t_type, int *pfitsheadsize) {
  int i, l, pos;
  char str[70];
  char *pc = (char *) ptr;
  
  if ((pos = fitsfind(fitsbuf, keyword)) < 0)
    fitsadd(&fitsbuf,keyword, NULL, pfitsheadsize);
  
  switch(type) {
  case H_INT:	sprintf(str, "%20d", (t_type==T_SHORT)?
			*(short *)ptr: *(int *)ptr);
    break;
    
  case H_FLOAT:	sprintf(str, "        %12.4f", (t_type==T_DOUBLE)? *(double *)ptr: *(float *)ptr);
    break;
    
  case H_EXPO:	sprintf(str, "    %16.9e", (t_type==T_DOUBLE)? *(double *)ptr: *(float *)ptr);
    break;
  case H_BOOL:	if (*(int *)ptr)
    sprintf(str, "                   T");
  else
    sprintf(str, "                   F");
    break;
  case H_STRING:	l = strlen(pc);
    for (i=0; i < l; i++) str[i] = pc[i];
    for (i=l; i < 69; i++) str[i] = ' ';
    str[69] = '\0';
    break;
  case H_COMMENT:	sprintf(str, "%69s", (char *)ptr);
    break;
  default:		error(EXIT_FAILURE,
			      "*FATAL ERROR*: Unknown FITS type in ",
			      "fitswrite()");
    break;
  }
  
  if ((pos = fitsfind(fitsbuf, keyword)) < 0) {
    return FAILURE;
  } else {
    fitsbuf += 80*pos+10;
    i = 0;
    while (str[i] != '\0') {
      fitsbuf[i] = str[i];
      i++;
    }
  }
  
  return SUCCESS;
}

/* 
 * PURPOSE: Update the fields 'fitshead' and 'fitsheadsize' of a FITSSTRUCT structure from the values 
 * of other fields of the FITSSTRUCT structure.
 *
 * INPUT: *field --- FITSSTRUCT, a fits header structure
 *
 */
void updatefitsheader(fitsstruct *field) {
  int status;
  // extern softinfo Soft;
  
  if (DEBUH) cout << "updatefitsheader entry " << endl;
  
  if (field->fitsheadsize < FBSIZE) field->fitsheadsize=FBSIZE;
  if (field->fitshead == NULL) {
    field->fitshead = creafitsheader();
    field->fitsheadsize=FBSIZE;
  }
  fitswrite(field->fitshead, "NAXIS   ", &(field->naxis), H_INT, T_INT,&(field->fitsheadsize));

  if (field->naxis > 0)
    fitswrite(field->fitshead, "NAXIS1  ", &(field->width), H_INT, T_INT,&(field->fitsheadsize));
  
  if (field->naxis > 1)
    fitswrite(field->fitshead, "NAXIS2  ", &(field->height), H_INT, T_INT,&(field->fitsheadsize));
  
  if (field->naxis > 2) 
    fitswrite(field->fitshead, "NAXIS3  ", &((field->TabAxis)[2]), H_INT, T_INT,&(field->fitsheadsize));
  fitswrite(field->fitshead, "BITPIX  ", &(field->bitpix), H_INT, T_INT,&(field->fitsheadsize));
  if (((field->bzero != 0.) || (field->bscale != 1.)) && (field->bscale != 0.)) {
    fitswrite(field->fitshead, "BSCALE  ",&(field->bscale), H_EXPO, T_DOUBLE,&(field->fitsheadsize));
  }
  if (ABS(field->bzero) > FLOAT_EPSILON)
    fitswrite(field->fitshead, "BZERO  ", &(field->bzero), H_EXPO, T_DOUBLE,&(field->fitsheadsize));
  
  if (ABS(field->crpixx)  > FLOAT_EPSILON)
    fitswrite(field->fitshead, "CRPIX1  ", &(field->crpixx), H_EXPO, T_DOUBLE,&(field->fitsheadsize));
  if (ABS(field->crpixy) > FLOAT_EPSILON)
    fitswrite(field->fitshead, "CRPIX2  ", &(field->crpixy), H_EXPO, T_DOUBLE,&(field->fitsheadsize));
  if (ABS((*field).TabRef[2]) > FLOAT_EPSILON)
    fitswrite(field->fitshead, "CRPIX3  ", &((field->TabRef)[2]), H_EXPO, T_DOUBLE,&(field->fitsheadsize));
  
  if (ABS(field->crvalx) > FLOAT_EPSILON)
    fitswrite(field->fitshead, "CRVAL1  ", &(field->crvalx), H_EXPO, T_DOUBLE,&(field->fitsheadsize));
  if (ABS(field->crvaly) > FLOAT_EPSILON)
    fitswrite(field->fitshead, "CRVAL2  ", &(field->crvaly), H_EXPO, T_DOUBLE,&(field->fitsheadsize));
  if (ABS((*field).TabValRef[2]) > FLOAT_EPSILON)
    fitswrite(field->fitshead, "CRVAL3  ", &((field->TabValRef)[2]), H_EXPO, T_DOUBLE,&(field->fitsheadsize));
  
  if (ABS(field->cdeltx) > FLOAT_EPSILON)
    fitswrite(field->fitshead, "CDELT1  ", &(field->cdeltx), H_EXPO, T_DOUBLE,&(field->fitsheadsize));
  if (ABS(field->cdelty) > FLOAT_EPSILON)
    fitswrite(field->fitshead, "CDELT2  ", &(field->cdelty), H_EXPO, T_DOUBLE,&(field->fitsheadsize));
  if (ABS((*field).TabStep[2]) > FLOAT_EPSILON)
    fitswrite(field->fitshead, "CDELT3  ", &((field->TabStep)[2]), H_EXPO, T_DOUBLE,&(field->fitsheadsize));
  
  if ((strlen(field->ctypex) > 1) && (strlen(field->ctypey) > 1)) {
    fitswrite(field->fitshead, "CTYPE1  ", 
	      (void*)  "          ", H_STRING, T_INT,&(field->fitsheadsize));
    fitswrite(field->fitshead, "CTYPE1  ", 
	      field->ctypex, H_STRING, T_INT,&(field->fitsheadsize));
    fitswrite(field->fitshead, "CTYPE2  ", 
	      (void*)  "          ", H_STRING, T_INT,&(field->fitsheadsize));
    fitswrite(field->fitshead, "CTYPE2  ", 
	      field->ctypey, H_STRING, T_INT,&(field->fitsheadsize));
  }
  
  //if (DEBUH) cout <<  Soft.banner() << endl;
  /* Add the soft copyright in the history field of the FITSSTRUCT structure */
  //	status = fitsaddhist_com(field,Soft.banner(),"HISTORY");
  
  if (DEBUH) cout << " field->origin = " << field->origin << endl;
  if (strlen(field->origin)  > 1) {
    status = fitsaddhist_com(field,field->origin,"HISTORY"); 
    field->origin = "";
  }
  
  if (DEBUH) cout << " field->rident = " << field->rident << endl;
  
  if (strlen(field->rident)  > 1) 
    status = fitsaddhist_com(field,field->rident,"HISTORY"); 
  if (DEBUH) cout << "5" << endl;
  
  status = fitswritehist(field);
  
  field->bytepix = ABS(field->bitpix) / 8;
  if (DEBUH) cout << "END" << endl;
}

/*
 * PURPOSE: Write down on the disk the FITS header enclosed in the field 'fitshead'
 * of a FITSSTRUCT structure. Before writing down on the disk, one must update
 * the header from the fields of the FITSSTRUCT structure.
 *
 * INPUT: *Header --- FITSSTRUCT, fits header structure
 *
 */
void writeimagehead(fitsstruct *Header) {
  /* Update the header hold in the field 'fitshead' */
  updatefitsheader(Header);
  
  
  /* Write down on the disk */
  QFWRITE(Header->fitshead,Header->fitsheadsize,Header->file,Header->filename);
}

/*
 * Read and convert input data stream in float.
 */
void  writedataf(FILE *file, char *filename, int bitpix, int Nelem, float *ptr, float bscale, float bzero) {
  int i,size=Nelem;
  char *bufdata;
  short *bufs;
  int *bufi;
  float *buff;
  double *bufd;
  float Val;
  long NbrByte = size * (ABS(bitpix) / 8);
  long FitsNbrByte = FBSIZE - (int) (NbrByte % FBSIZE);
 
// cout << " NbrByte = " << NbrByte << " FitsNbrByte = " << FitsNbrByte << endl;
  
   // if (!(bufdata=new char [FitsNbrByte])) 
   //     error(EXIT_FAILURE, "*Error*: Not enough memory in ",  "writedataf()");
#if BYTESWAPPED == 1
    bufdata = (char*)malloc((size_t)NbrByte);
#else
    if (bitpix != BP_FLOAT)
               bufdata = (char*)malloc((size_t)NbrByte);
    else bufdata = (char *) ptr;
#endif
 
    char *FitsBit = new char [FitsNbrByte];
    for (i = 0; i < FitsNbrByte; i++) FitsBit[i] = 0;
     	       
    bufs = (short *) bufdata;
    bufi = (int *) bufdata;
    buff = (float *) bufdata;
    bufd = (double *) bufdata;

// cout << "writedataf" << endl << "BITPIX = " << bitpix  << endl;

    switch(bitpix) {
      case BP_BYTE:	for (i=0; i< size; i++) {
                           Val = (ptr[i] - bzero) / bscale;
                           //if (Val > 127) Val = 127;
                           //if (Val < -128) Val = -128;
                           bufdata[i] = (char) Val;
			}
			break;
      case BP_SHORT:
			for (i=0; i< size; i++) {
                           Val = (ptr[i] - bzero) / bscale;
                           if (Val > 0) bufs[i] = (int) (Val+0.5);
                           else  bufs[i] = (int) (Val-0.5);
 			}
/* the 3 following lines come from CFITSIO */
#if BYTESWAPPED == 1
    ffswap2((short int *)bufdata, size); /* reverse order of bytes in each value */
#endif
 			break;
      case BP_INT:
			for (i=0; i< size; i++) {
                           Val = (float)((double)(ptr[i] - bzero) / bscale);
                           if (Val > 0) bufi[i] = (int) (Val+0.5);
                           else  bufi[i] = (int) (Val-0.5);
                        }
/* the 3 following lines come from CFITSIO */
#if BYTESWAPPED == 1
     ffswap4((INT32BIT *) bufdata, size); /* reverse order of bytes in each value */
#endif
			break;
      case BP_FLOAT:
			for (i=0; i< size; i++)
                           buff[i] = (ptr[i] - bzero) / bscale;
#if BYTESWAPPED == 1
    ffswap4((INT32BIT *) bufdata, size); /* reverse order of bytes in each value */
#endif
			break;
      case BP_DOUBLE:
			for (i=0; i< size; i++)
                           bufd[i] = (double)((ptr[i] - bzero) / bscale);
#if BYTESWAPPED == 1
    ffswap8((double *)bufdata, size); /* reverse order of bytes in each value */
#endif
			break;
      default:		error(EXIT_FAILURE,
				"*FATAL ERROR*: unknown BITPIX type in ",
				"writedata()");
			break;
      }
//  cout << "Nelem = " << Nelem << endl;
//  cout << "bitpix = " << bitpix << endl;
//  cout << "NbrByte = " << NbrByte << endl;
  
  QFWRITE(bufdata, NbrByte,file, filename);
  QFWRITE(FitsBit, FitsNbrByte,file, filename);

#if BYTESWAPPED == 1
  free((char *) bufdata);
#else
  if (bitpix != BP_FLOAT) free((char *) bufdata);
#endif

  delete [] FitsBit;
  // delete bufdata;
}

void  writedataf(fitsstruct *field, float *ptr) {

// cout << "filename = " << field->filename << endl;
// cout << "npix = " << field->npix << endl;
// cout << "bscale = " << field->bscale << endl;
// cout << "bzero = " << field->bzero << endl;
// cout << "bitpix = " << field->bitpix << endl;

    writedataf(field->file, field->filename, field->bitpix,field->npix, ptr, (float)field->bscale, (float)field->bzero);
}

/*
 * search for a FITS keyword in a fits header of nblock blocks.
 */
char *fitsnfind(char *fitsbuf, char *str, int nblock) {
  int i;

  for (i=0;i<36*nblock;i++)
      if (!strncmp(&fitsbuf[80*i], str, strlen(str)))
              return &fitsbuf[80*i];

  return (char *)NULL;
}

/*
 * read data from the FITS-file header 
 */
char  *readfitshead( FILE *file, char *filename, int *nblock ) {
   int     n;
   char    *buf;
   
   if (!(buf= new char[FBSIZE]))
      error(EXIT_FAILURE, "*Error*: Not enough memory in ", "readfitshead()");
   
   /*Find the number of FITS blocks of the header while reading it */
   QFREAD(buf, FBSIZE, file, filename);
   
   if ( strncmp(buf, "SIMPLE  ", 8) && strncmp(buf, "XTENSION", 8) )
      error(EXIT_FAILURE, filename, " is not a FITS file!");
   
   for (n=1; !fitsnfind(buf,"END     ", n); n++) {
      if (!(buf=(char *)realloc(buf, (size_t)(FBSIZE*(n+1)))))
	 error(EXIT_FAILURE, "*Error*: Not enough memory in ", "readfitshead()");
      QFREAD(&buf[FBSIZE*n], FBSIZE, file, filename);
   }
   *nblock = n;
   return  buf;
}

/*
 * Extract some data from the FITS-file header.
 */
void readimagehead( fitsstruct *field ) {
  int		i, n;
  char		*buf, st[80], *point;
  
  buf = readfitshead(field->file, field->filename, &n);
  field->naxis = FITSTOI("NAXIS   ", 0);
  if(field->naxis < 1)
    error(EXIT_FAILURE, field->filename, " NAXIS is not equal to 1,2, or 3 !");
  
  field->bitpix = FITSTOI("BITPIX  ", 0);
  if (field->bitpix != BP_BYTE
      && field->bitpix != BP_SHORT
      && field->bitpix != BP_INT
      && field->bitpix != BP_FLOAT
      && field->bitpix != BP_DOUBLE)
    error(EXIT_FAILURE, "Sorry, I don't know that kind of data.", "");
  
  field->bitpix = FITSTOI("BITPIX  ", 0);
  if (field->bitpix != BP_BYTE
      && field->bitpix != BP_SHORT
      && field->bitpix != BP_INT
      && field->bitpix != BP_FLOAT
      && field->bitpix != BP_DOUBLE)
    error(EXIT_FAILURE, "Sorry, I don't know that kind of data.", "");
  
  field->bytepix = (field->bitpix>0?field->bitpix:-field->bitpix)>>3;
  if (field->naxis > 0) {
    field->width =  FITSTOI("NAXIS1  ", 0);
    field->crpixx = FITSTOF("CRPIX1  ", 0.);
    field->crvalx = FITSTOF("CRVAL1  ", 0.0);
    field->cdeltx = FITSTOF("CDELT1  ", 0.0);
    (field->TabAxis)[0] = field->width;
    (field->TabRef)[0] = field->crpixx;
    (field->TabValRef)[0] = field->crvalx;
    (field->TabStep)[0] = field->cdeltx;
    field->npix = field->width;
    field->crotax = FITSTOF("CROTA1  ", 0.0);
    FITSTOS("CTYPE1 ", field->ctypex, "       ");
  }
  if (field->naxis > 1) {
    field->height = FITSTOI("NAXIS2  ", 0);
    field->crpixy = FITSTOF("CRPIX2  ", 0.);
    field->crvaly = FITSTOF("CRVAL2  ", 0.0);
    field->cdelty = FITSTOF("CDELT2  ", 0.0);
    (field->TabAxis)[1] = field->height;
    (field->TabRef)[1] = field->crpixy;
    (field->TabValRef)[1] = field->crvaly;
    (field->TabStep)[1] = field->cdelty;
    field->npix = field->width*field->height;
    field->crotay = FITSTOF("CROTA2  ", 0.0);
    FITSTOS("CTYPE2 ", field->ctypey, "       ");
  }
  if (field->naxis > 2) {
    (field->TabAxis)[2] = FITSTOI("NAXIS3  ", 0);
    (field->TabRef)[2] = FITSTOF("CRPIX3  ", 0.);
    (field->TabValRef)[2] = FITSTOF("CRVAL3  ", 0.0);
    (field->TabStep)[2] = FITSTOF("CDELT3  ", 0.0);
    field->npix = field->npix*(field->TabAxis)[2];
  }
  
  field->bscale = FITSTOF("BSCALE  ", 1.0);
  field->bzero = FITSTOF("BZERO   ", 0.0);
  field->epoch = FITSTOF("EPOCH   ", 0.0);
  
  //FITSTOS("OBJECT  ", field->ident, "Unknown");
  
  field->fitshead = buf;
  field->fitsheadsize = n*FBSIZE;
  field->origin = "\0";
  strcpy(field->rident, "");
  
  /* HISTORY & COMMENT fields */
  //field->hist_size = fitsreadhist(field->fitshead,field->fitsheadsize,"HISTORY",&field->history);
  //field->com_size = fitsreadhist(field->fitshead,field->fitsheadsize,"COMMENT",&field->comment);
  
}

void  readdataf(FILE *file, char *filename, int bitpix, int Nelem, float *ptr, float bscale, float bzero) {
  /* read and convert input data stream in float. */
  int i,size=Nelem;
  char *bufdata;
  // float *ptrf;
  long NbrByte = size * (ABS(bitpix) / 8);
  
  //if (!(bufdata=new char [NbrByte]))
  //     error(EXIT_FAILURE, "*Error*: Not enough memory in ",  "readdata()");
  if (bitpix != BP_FLOAT)
    bufdata = (char*)malloc((size_t)NbrByte);
  else bufdata = (char *) ptr;
  QFREAD(bufdata, NbrByte, file, filename);
  
  /*    
	#if BYTESWAPPED == 1
	cout << "MACHINE = " << MACHINE <<  endl;
	cout << "BYTESWAPPED = True" << endl;
	#endif
  */
  switch(bitpix) {
  case BP_BYTE: 
    for (i=0; i< size; i++) ptr[i] = ((unsigned char *)bufdata)[i]*bscale+bzero;
    break;
    
  case BP_SHORT:    
    /* the 3 following lines come from CFITSIO */
#if BYTESWAPPED == 1
    ffswap2((short int *)bufdata, size);    /* reverse order of bytes in each bufdata */
#endif
    
    for (i=0; i< size; i++)
      ptr[i] = ((short int *)bufdata)[i]*bscale+bzero;
    break;
    
  case BP_INT:      
    /* the 3 following lines come from CFITSIO */
#if BYTESWAPPED == 1
    ffswap4((INT32BIT *) bufdata, size);  /* reverse order of bytes in each value */
#endif
    for (i=0; i<size; i++)
      ptr[i] = ((long int *)bufdata)[i]*bscale+bzero;
    break;
  case BP_FLOAT:
#if BYTESWAPPED == 1
    ffswap4((INT32BIT *) bufdata, size); /* reverse order of bytes in each value */
#endif
    for (i=0; i<size; i++) {
	ptr[i] = ((float *) bufdata)[i]*bscale+bzero;
      }
    break;
  case BP_DOUBLE:
#if BYTESWAPPED == 1
    ffswap8((double *)bufdata, size);   /* reverse order of bytes in each value */
#endif
    for (i=0; i<size; i++)
      ptr[i] = ((float *)bufdata)[i]*bscale+bzero;
    break;
    
  default:		error(EXIT_FAILURE,
			      "*FATAL ERROR*: unknown BITPIX type in ",
			      "readdata()");
    break;
  }
  if (bitpix != BP_FLOAT) free((char *) bufdata);
  // delete bufdata;
  return;
}

#endif
