/* -*- mode: c++; c-basic-offset: 3 -*-
 *
 * PANDORE (PANTHEON Project)
 *
 * GREYC IMAGE
 * 6 Boulevard Marchal Juin
 * F-14050 Caen Cedex France
 *
 * This file is free software. You can use it, distribute it
 * and/or modify it. However, the entire risk to the quality
 * and performance of this program is with you.
 *
 * Please email any bugs, comments, and/or additions to this file to:
 * pandore@greyc.ensicaen.fr
 * http://www.greyc.ensicaen.fr/EquipeImage/Pandore/
 */

/**
 * @author Jalal Fadili - 02-02-2005
 * @author Christophe Glinel - 20-07-2005
 */

/**
 * @file ildctiv.cpp
 *

 * Inverse Local Discrete Cosine Transform IV
 * The ILDCTIV is only implemented with 2D gray
 * images.
 */
 
/**
 * Transforme discrte inverse en cosinus locaux IV (ILDCTIV) (voir aussi les transforme en paquets de cosinus).
 * La DCT implmente ici ne traite que le cas des images
 * 2D de niveaux de gris.
 */

#include "dctutils.cpp"


/**
 * ILDCTIV
 * Integer 2D gray image. (Img2duc)
 * Images MUST have their dimensions all powers of 2 -> reshape.
 * The dimensions of the input and output must be the same.
 * Inputs :
 *	ims   : The input image.
 *	imd   : The output image
 *	ww    : The width of the window.
 *	lfign : ...
 */
Errc ILDCTIV(Img2duc &ims,Img2dsf &imd,Ulong ww,Float lfign){

   Long x,y;
   Long width   = imd.Width();
   Long height  = imd.Height();
   Double *data = new Double[MAX(width,height)];
	
   int nbox  = (int)floor((Double)MIN(width,height)/(Double)ww);
   int lfcut = (int)floor((Double)ww*lfign);
	
   if(lfign){ 
      // This is sometimes useful for texture separation.
      for ( int d1 = 0 ; d1 < nbox ; d1++)
	 for ( int d2 = 0 ; d2 < nbox ; d2++)
	    for ( y = 0 ; y < lfcut ; y++)
	       for ( x = 0 ; x < lfcut ; x++)
		  ims[d1*ww+y][d2*ww+x] = 0;
   }
	
   /* Go through rows. */
   for ( y = 0 ; y < height ; y++){
		
      for( x = 0 ; x < width ; x++)
	 data[x] = (Double)ims[y][x];
		
      ildct_iv1D(data, ww, width);
		
      for( x = 0 ; x < width ; x++)
	 imd[y][x] = (Img2dsf::ValueType)data[x];
   }
	
   /* Go through columns. */
   for ( x = 0 ; x < width ; x++){
		 
      for( y = 0 ; y < height ; y++ )
	 data[y] = (Double)imd[y][x];
		
      ildct_iv1D(data, ww, height);
		
      for( y = 0 ; y < height ; y++ )
	 imd[y][x] = (Img2dsf::ValueType)data[y];
   }  
	
   delete []data;
   return SUCCESS;
}

/**
 * ILDCTIV
 * Integer 2D gray image. (Img2dsl)
 * Images MUST have their dimensions all powers of 2 -> reshape.
 * The dimensions of the input and output must be the same.
 * Inputs :
 *	ims   : The input image.
 *	imd   : The output image
 *	ww    : The width of the window.
 *	lfign : ...
 */
Errc ILDCTIV(Img2dsl &ims,Img2dsf &imd,Ulong ww,Float lfign){

   Long x,y;
   Long width   = imd.Width();
   Long height  = imd.Height();
   Double *data = new Double[MAX(width,height)];
	
   int nbox  = (int)floor((Double)MIN(width,height)/(Double)ww);
   int lfcut = (int)floor((Double)ww*lfign);
	
   if(lfign){ 
      // This is sometimes useful for texture separation.
      for ( int d1 = 0 ; d1 < nbox ; d1++)
	 for ( int d2 = 0 ; d2 < nbox ; d2++)
	    for ( y = 0 ; y < lfcut ; y++)
	       for ( x = 0 ; x < lfcut ; x++)
		  ims[d1*ww+y][d2*ww+x] = 0;
   }
	
   /* Go through rows. */
   for ( y = 0 ; y < height ; y++){
		
      for( x = 0 ; x < width ; x++)
	 data[x] = (Double)ims[y][x];
		
      ildct_iv1D(data, ww, width);
		
      for( x = 0 ; x < width ; x++)
	 imd[y][x] = (Img2dsf::ValueType)data[x];
   }
	
   /* Go through columns. */
   for ( x = 0 ; x < width ; x++){
		 
      for( y = 0 ; y < height ; y++ )
	 data[y] = (Double)imd[y][x];
		
      ildct_iv1D(data, ww, height);
		
      for( y = 0 ; y < height ; y++ )
	 imd[y][x] = (Img2dsf::ValueType)data[y];
   }  
	
   delete []data;
   return SUCCESS;
}

/**
 * ILDCTIV
 * Integer 2D gray image. (Img2dsf)
 * Images MUST have their dimensions all powers of 2 -> reshape.
 * The dimensions of the input and output must be the same.
 * Inputs :
 *	ims   : The input image.
 *	imd   : The output image
 *	ww    : The width of the window.
 *	lfign : ...
 */
Errc ILDCTIV(Img2dsf &ims,Img2dsf &imd,Ulong ww,Float lfign){

   Long x,y;
   Long width   = imd.Width();
   Long height  = imd.Height();
   Double *data = new Double[MAX(width,height)];
	
   int nbox  = (int)floor((Double)MIN(width,height)/(Double)ww);
   int lfcut = (int)floor((Double)ww*lfign);
	
   if(lfign){ 
      // This is sometimes useful for texture separation.
      for ( int d1 = 0 ; d1 < nbox ; d1++)
	 for ( int d2 = 0 ; d2 < nbox ; d2++)
	    for ( y = 0 ; y < lfcut ; y++)
	       for ( x = 0 ; x < lfcut ; x++)
		  ims[d1*ww+y][d2*ww+x] = 0;
   }
	
   /* Go through rows. */
   for ( y = 0 ; y < height ; y++){
		
      for( x = 0 ; x < width ; x++)
	 data[x] = (Double)ims[y][x];
		
      ildct_iv1D(data, ww, width);
		
      for( x = 0 ; x < width ; x++)
	 imd[y][x] = (Img2dsf::ValueType)data[x];
   }
	
   /* Go through columns. */
   for ( x = 0 ; x < width ; x++){
		 
      for( y = 0 ; y < height ; y++ )
	 data[y] = (Double)imd[y][x];
		
      ildct_iv1D(data, ww, height);
		
      for( y = 0 ; y < height ; y++ )
	 imd[y][x] = (Img2dsf::ValueType)data[y];
   }  
	
   delete []data;
   return SUCCESS;
}

/**
 * ILDCTIV
 * Floating-point 3D gray image (Img3dsf)
 * Images MUST have their dimensions all powers of 2 -> reshape.
 * The dimensions of the input and output must be the same.
 * Inputs :
 *	ims   : The input image.
 *	imd   : The output image
 *	ww    : The width of the window.
 */
Errc ILDCTIV(Img3dsf &ims,Img3dsf &imd,Ulong ww){

   Long x,y,z;
   Long width   = imd.Width();
   Long height  = imd.Height();
   Long depth   = imd.Depth();
   Img2dsf sframe(height,width);
   Img2dsf dframe;
   Double *data = new Double[depth];
   
   for ( y = 0 ; y < height ; y++){	
      for( x = 0 ; x < width ; x++){
        for( z = 0 ; z < depth ; z++)
	  data[z] = (Double)ims[z][y][x];
		
        ildct_iv1D(data, ww, depth);
	
	for( z = 0 ; z < depth ; z++)
	  imd[z][y][x] = (Img3dsf::ValueType)data[z];
      }
   }
   
   /* Go through frames. */
   for ( z = 0 ; z < depth ; z++){
   	memcpy(sframe.Vector(),imd.Vector()+z*height*width,height*width*sizeof(Float));
	
   	dframe.New(height,width,&imd[z][0][0]);
	
	ILDCTIV(sframe,dframe,ww,0);
   }
   
	
   return SUCCESS;
}

#ifdef MAIN
#define USAGE	"USAGE : %s window_width [im_in|-] [im_out|-]"
#define PARC	1
#define FINC	1
#define FOUTC	1
#define MASK	0
int main( int argc, char *argv[] ) {
   Errc result;              // The result code of the execution.
   Pobject* mask;            // The region map.
   Pobject* objin[FINC+1];   // The input objects;
   Pobject* objs[FINC+1];    // The source objects masked.
   Pobject* objout[FOUTC+1]; // The output object.
   Pobject* objd[FOUTC+1];   // The result object of the execution.
   char* parv[PARC+1];       // The input parameters.

   ReadArgs(argc,argv,PARC,FINC,FOUTC,&mask,objin,objs,objout,objd,parv,USAGE,MASK);
if (objs[0]->Type() == Po_Img2duc){

   Img2duc * const ims=(Img2duc*)objs[0];
   Check(*ims);
   objd[0] = new Img2dsf(ims->Size());
   Img2dsf * const imd = (Img2dsf*)objd[0];
   result = ILDCTIV(*ims,*imd,atol(parv[0]),0);
   goto end;
   
}
else 
if (objs[0]->Type() == Po_Img2dsl){

   Img2dsl * const ims=(Img2dsl*)objs[0];
   Check(*ims);
   objd[0] = new Img2dsf(ims->Size());
   Img2dsf * const imd = (Img2dsf*)objd[0];
   result = ILDCTIV(*ims,*imd,atol(parv[0]),0);
   goto end;
   
}
else 
if (objs[0]->Type() == Po_Img2dsf){

   Img2dsf * const ims=(Img2dsf*)objs[0];
   Check(*ims);
   objd[0] = new Img2dsf(ims->Size());
   Img2dsf * const imd = (Img2dsf*)objd[0];
   result = ILDCTIV(*ims,*imd,atol(parv[0]),0);
   goto end;
   
}
else 
if (objs[0]->Type() == Po_Img3dsf){

   Img3dsf * const ims=(Img3dsf*)objs[0];
   Check(*ims);
   objd[0] = new Img3dsf(ims->Size());
   Img3dsf * const imd = (Img3dsf*)objd[0];
   result = ILDCTIV(*ims,*imd,atol(parv[0]));
   goto end;
   
}
else 
  {
     PrintErrorFormat(objin,FINC);
     result = FAILURE;
  }	

end:
  if (result)
	WriteArgs(argc,argv,PARC,FINC,FOUTC,&mask,objin,objs,objout,objd,MASK);
  Exit(result);
  return 0;
}
#endif
