/* -*- mode: c++; c-basic-offset: 3 -*-
 *
 * PANDORE (PANTHEON Project)
 *
 * GREYC IMAGE
 * 6 Boulevard Marchal Juin
 * F-14050 Caen Cedex France
 *
 * This file is free software. You can use it, distribute it
 * and/or modify it. However, the entire risk to the quality
 * and performance of this program is with you.
 *
 * Please email any bugs, comments, and/or additions to this file to:
 * pandore@greyc.ensicaen.fr
 * http://www.greyc.ensicaen.fr/EquipeImage/Pandore/
 */

/**
 * @author Jalal Fadili - 02-02-2005
 */

/**
 * @file wutils.cpp
 *
 */

#ifndef WUTILS_CC
#define WUTILS_CC
#include "utils.cpp"

/**
 * A dictionary is a collection of transforms.
 * For more details see the dictionary.cpp
 */
class Dictionary {
public :
   
   // Number of transforms in the dictionary.
   Ulong NbTransforms;
   
   // Pointer to a collection of collections.
   Collection **Dict;
   
   // Components of a transform.
   char *NameofTransform;
   Float Par1;
   Float *Par2;
   Ulong Size;
   Float Par3;

   Dictionary(Collection & c) {
      NbTransforms = c.GETPARRAYSIZE("Dict",Collection);
      Dict = c.GETPARRAY("Dict",Collection);
      NameofTransform = Dict[0]->GETARRAY("NameOfDict",char);
      Par1 = Dict[0]->GETVALUE("PAR1",Float);
      Par2 = Dict[0]->GETARRAY("PAR2",Float);
      Size = Dict[0]->GETARRAYSIZE("PAR2", Float);
      Par3 = Dict[0]->GETVALUE("PAR3",Float);
   }
   
   /**
    * To know the number of transforms in the dictionary.
    */
   Ulong GetNbTransforms()const { return NbTransforms;}
   
   /**
    * Gives the name of the transform.
    */
   char *  GetNameofTransform() { return NameofTransform; }
	
   /**
    * Gives the first parameter.
    */
   Float GetPar1()const { return Par1;}
	
   /**
    * Gives the second.
    */
   Float * GetPar2() { return Par2;}
	
   /**
    * Gives the length of the QMF.
    */
   Ulong GetSize()const { return Size;}
	
   /**
    * Gives the third.
    */
   Float GetPar3()const { return Par3;}
	
   /**
    * To display the dictionary.
    */
   friend std::ostream & operator<<(std::ostream &out,Dictionary & d);
	
   /**
    * To display one transform of the dictionary.
    */
   void Display() {
      std::cout<<NameofTransform<< " ";
      std::cout<<Par1<<" ";
      std::cout<<Par2[0]<<" ";
      std::cout<<Size<<" ";
      std::cout<<Par3<<std::endl;
   }
	
   /**
    * This operator allows us to give the good transform
    * to fat operator and fst operator.
    * Inputs :
    *	i : The index of the wanted transform.
    */	
   Dictionary  & operator[](Ulong i) {
      if(i >= NbTransforms) {
	 std::cerr<< "The number of transforms is " << NbTransforms << std::endl;
	 Exit(FAILURE);
      }
      
      NameofTransform = Dict[i]->GETARRAY("NameOfDict",char);
      Par1 = Dict[i]->GETVALUE("PAR1",Float);
      Par2 = Dict[i]->GETARRAY("PAR2",Float);
      Size = Dict[i]->GETARRAYSIZE("PAR2", Float);
      Par3 = Dict[i]->GETVALUE("PAR3",Float);
      return *this;	
   }
};

/**
 * To display the dictionary.
 */
std::ostream & operator<<(std::ostream &out,Dictionary & d) {
   Ulong i;
   for ( i = 0 ; i < d.GetNbTransforms() ; i++) {
	 out<<d.Dict[i]->GETARRAY("NameOfDict",char)<< " ";
	 out<<d.Dict[i]->GETVALUE("PAR1",Float)<<" ";
	 out<<d.Dict[i]->GETARRAY("PAR2",Float)[0]<<" ";
	 out<<d.Dict[i]->GETARRAYSIZE("PAR2", Float)<<" ";
	 out<<d.Dict[i]->GETVALUE("PAR3",Float)<<std::endl;
      }
   return out;
}

/**
 * We define what is a QMF.
 */
class Qmf : public Collection {

private :
   // Name of the QMF.
   char * NameQmf;
	
   static const int Max_Length = 15;
	
   // Coefficients of the QMF.
   Float *Coeffs;
	
   // The length of the vector
   int Size;
	
   /**
    * Save the data into the collection.
    * Inputs :
    * 	c : The collection to fill.
    */
   Errc Save(Collection & c) {
      c.SETARRAY(NameQmf, Float, Coeffs, Size);
      c.SETARRAY("filtre", char, NameQmf, Max_Length); 
      return SUCCESS;
   }
	
protected :
public :
	
   /**
    * Constructor
    */
   Qmf():Collection(),NameQmf(0){}
	
   /**
    * Destructor
    */
   ~Qmf(){}
	
   /**
    * To create a QMF from a dictionary.
    * Inputs :
    *	d : The dictionary.
    */
   Qmf(Dictionary & d):Collection(),NameQmf(0) {
      Coeffs     = d.GetPar2();
      Size       = d.GetSize();
   }
	
   /**
    * To create a QMF from a collection
    * Inputs :
    *	c : The collection.
    */	
   Qmf(Collection &c):Collection(),NameQmf(0) {
      NameQmf    = c.GETARRAY("filtre", char);
      Coeffs     = c.GETARRAY(NameQmf, Float);
      Size       = c.GETARRAYSIZE(NameQmf, Float);
   }
	
   /**
    * To create a QMF with an existing QMF which is saved on the disk.
    * Inputs :
    * 	path : The QMF to load.
    */
   Qmf(char *path) {
      if( Collection::LoadFile(path) == FAILURE ) {
	 std::cerr<< "The QMF file " << path <<" cannot be loaded." <<std::endl;
	 Exit(FAILURE);
      }
      NameQmf    = this->GETARRAY("filtre", char);
      Coeffs     = this->GETARRAY(NameQmf, Float);
      Size       = this->GETARRAYSIZE(NameQmf, Float);
   }

   /**
    * We build a QMF filter with the name of
    * the wavelet and the order.
    *
    * References
    * The books by Daubechies and Wickerhauser.
    */
   Qmf(char  * wavelet, int order):Collection(),NameQmf(0) {
	
      int size = 0;
      //int s = -1;
      //int k;
      Float *f = 0;
      Float normf;
	
      if(!strcmp(wavelet, "beylkin")) {
	 size = 18;
	 static Float h[19] = { 0.0f, .099305765374f, .424215360813f, .699825214057f, .449718251149f, -.110927598348f, -.264497231446f, 
				.026900308804f, .155538731877f, -.017520746267f, 
				-.088543630623f, .019679866044f, .042916387274f, 
				-.017460408696f, -.014365807969f, .010040411845f, 
				.001484234782f, -.002736031626f, .000640485329f };
	 f = h;
      } else if(!strcmp(wavelet, "coiflet")) {
	 size = order*6;

	 switch(order){
	 case 1 :
	    {
	       static Float h[7] = { 0.0f, 	.038580777748f, -.126969125396f, -.077161555496f, 
				     .607491641386f, .745687558934f, .226584265197f };
	       f = h; 
	    }
	    break;
	 case 2 :
	    {
	       static Float h[13] = { 0.0f, 	.016387336463f, -.041464936782f, -.067372554722f, 
				      .386110066823f, .812723635450f, .417005184424f, 
				      -.076488599078f, -.059434418646f, .023680171947f, 
				      .005611434819f, -.001823208871f, -.000720549445f };
	       f = h;
	    }
	    break;
	 case 3 :
	    {
	       static Float h[19] = { 0.0f, 	-.003793512864f, .007782596426f, .023452696142f, 
				      -.065771911281f, -.061123390003f, .405176902410f, 
				      .793777222626f, .428483476378f, -.071799821619f, 
				      -.082301927106f, .034555027573f, .015880544864f, 
				      -.009007976137f, -.002574517688f, .001117518771f, 
				      .000466216960f, -.000070983303f, -.000034599773f };
	       f = h;
	    }
	    break;
	 case 4 :
	    {
	       static Float h[25] = { 0.0f, 	.000892313668f, -.001629492013f, -.007346166328f, 
				      .016068943964f, .026682300156f, -.081266699680f, 
				      -.056077313316f, .415308407030f, .782238930920f, 
				      .434386056491f, -.066627474263f, -.096220442034f, 
				      .039334427123f, .025082261845f, -.015211731527f, 
				      -.005658286686f, .003751436157f, .001266561929f, 
				      -.000589020757f, -.000259974552f, .000062339034f, 
				      .000031229876f, -.000003259680f, -.000001784985f };
	       f = h;
	    }
	    break;
	 case 5 :
	    {
	       static Float h[31] = { 0.0f, 	-.000212080863f, .000358589677f, .002178236305f, 
				      -.004159358782f, -.010131117538f, .023408156762f, 
				      .028168029062f, -.091920010549f, -.052043163216f, 
				      .421566206729f, .774289603740f, .437991626228f, 
				      -.062035963906f, -.105574208706f, .041289208741f, 
				      .032683574283f, -.019761779012f, -.009164231153f, 
				      .006764185419f, .002433373209f, -.001662863769f, 
				      -.000638131296f, .000302259520f, .000140541149f, 
				      -.000041340484f, -.000021315014f, .000003734597f, 
				      .000002063806f, -.000000167408f, -.000000095158f };
	       f = h;
	    }
	    break;
	 default :
	    std::cerr<<"This order is not implemented "<<std::endl;
	    Exit( FAILURE);
	 }
      } else if(!strcmp(wavelet, "daubechies")){
			
	 size = order;
	 switch(order){
				
	 case 4:
	    {
	       static Float h[5] = { 0.0f,  .482962913145f, .836516303738f, .224143868042f, -.129409522551f };
	       f = h;
	    }
	    break;
	 case 6 :
	    {	
	       static Float h[7] = { 0.0f, 	.332670552950f, .806891509311f, .459877502118f, 
				     -.135011020010f, -.085441273882f, .035226291882f };
	       f = h;
	    }
	    break;
	 case 8 :
	    {
	       static Float h[9] = { 0.0f, 	.230377813309f, .714846570553f, .630880767930f,
				     -.027983769417f, -.187034811719f,	.030841381836f, 
				     .032883011667f, -.010597401785f };
	       f = h;
	    }
	    break;
	 case 10:
	    {
	       static Float h[11] = { 0.0f, 	.160102397974f, .603829269797f, .724308528438f, 
				      .138428145901f, -.242294887066f, -.032244869585f, 
				      .077571493840f, -.006241490213f, -.012580751999f, 
				      .003335725285f };
	       f = h;
	    }
	    break;
	 case 12:
	    {
	       static Float h[13] = { 0.0f, 	.111540743350f, .494623890398f, .751133908021f, 
				      .315250351709f, -.226264693965f, -.129766867567f, 
				      .097501605587f, .027522865530f, -.031582039317f, 
				      .000553842201f, .004777257511f, -.001077301085f };
	       f = h; 
	    }
	    break;
	 case 14 :
	    {
	       static Float h[15] = { 0.0f, 	.077852054085f, .396539319482f, .729132090846f, 
				      .469782287405f, -.143906003929f, -.224036184994f, 
				      .071309219267f, .080612609151f, -.038029936935f, 
				      -.016574541631f, .012550998556f, .000429577973f, 
				      -.001801640704f, .000353713800f };
	       f = h;
	    }
	    break;
	 case 16 :
	    {
	       static Float h[17] = { 0.0f, 	.054415842243f, .312871590914f, .675630736297f, 
				      .585354683654f, -.015829105256f, -.284015542962f, 
				      .000472484574f, .128747426620f, -.017369301002f, 
				      -.044088253931f, .013981027917f, .008746094047f, 
				      -.004870352993f, -.000391740373f, .000675449406f, 
				      -.000117476784f };
	       f = h;
	    }
	    break;
	 case 18 : 
	    {
	       static Float h[19] = { 0.0f, 	.038077947364f, .243834674613f, .604823123690f, 
				      .657288078051f, .133197385825f, -.293273783279f, 
				      -.096840783223f, .148540749338f, .030725681479f, 
				      -.067632829061f, .000250947115f, .022361662124f, 
				      -.004723204758f, -.004281503682f, .001847646883f, 
				      .000230385764f, -.000251963189f, .000039347320f };
	       f = h;
	    }
	    break;
	 case 20 :
	    {	
	       static Float h[21] = { 0.0f, 	.026670057901f, .188176800078f, .527201188932f, 
				      .688459039454f, .281172343661f, -.249846424327f, 
				      -.195946274377f, .127369340336f, .093057364604f, 
				      -.071394147166f, -.029457536822f, .033212674059f, 
				      .003606553567f, -.010733175483f, .001395351747f, 
				      .001992405295f, -.000685856695f, -.000116466855f, 
				      .000093588670f, -.000013264203f };
	       f = h;
	    }
	    break;
	 default :
	    std::cerr<<"This order is not implemented "<<std::endl;
	    Exit( FAILURE);
	 }
      } else if(!strcmp(wavelet, "symmlet")){
			
	 size = order*2;
	 switch(order){
			
	 case 4 :
	    {
	       static Float h[9] = { 0.0f,-.107148901418f, -.041910965125f, .703739068656f, 
				     1.136658243408f, .421234534204f, -.140317624179f, 
				     -.017824701442f, .045570345896f };
	       f = h;
	    }
	    break;
	 case 5 :
	    {	static Float h[11] = { 0.0f, 	.038654795955f, .041746864422f, -.055344186117f, 
				       .281990696854f, 1.023052966894f, .896581648380f, 
				       .023478923136f, -.247951362613f, -.029842499869f, 
				       .027632152958f };
	    f = h;
	    }
	    break;
	 case 6:
	    {
	       static Float h[13] = { 0.0f, 	.021784700327f, .004936612372f, -.166863215412f, 
				      -.068323121587f, .694457972958f, 1.113892783926f, 
				      .477904371333f, -.102724969862f, -.029783751299f, 
				      .063250562660f, .002499922093f, -.011031867509f };
	       f = h;
	    }
	    break;
	 case 7 :
	    {
	       static Float h[15] = { 0.0f, 	.003792658534f, -.001481225915f, -.017870431651f, 
				      .043155452582f, .096014767936f, -.070078291222f, 
				      .024665659489f, .758162601964f, 1.085782709814f, 
				      .408183939725f, -.198056706807f, -.152463871896f, 
				      .005671342686f, .014521394762f };
	       f = h;
	    }
	    break;
	 case 8 :
	    {
	       static Float h[17] = { 0.0f, 	.002672793393f, -.000428394300f, -.021145686528f, 
				      .005386388754f, .069490465911f, -.038493521263f, 
				      -.073462508761f, .515398670374f, 1.099106630537f, 
				      .680745347190f, -.086653615406f, -.202648655286f, 
				      .010758611751f, .044823623042f, -.000766690896f,  
				      -.004783458512f };
	       f = h;
	    }
	    break;
	 case 9 :
	    {
	       static Float h[19] = { 0.0f, 	.001512487309f, -.000669141509f, -.014515578553f, 
				      .012528896242f, .087791251554f, -.025786445930f, 
				      -.270893783503f, .049882830959f, .873048407349f, 
				      1.015259790832f, .337658923602f, -.077172161097f, 
				      .000825140929f, .042744433602f, -.016303351226f, 
				      -.018769396836f, .000876502539f, .001981193736f };
	       f = h;
	    }
	    break;
	 case 10 :
	    {
	       static Float h[21] = { 0.0f, 	.001089170447f, .000135245020f, -.012220642630f, 
				      -.002072363923f, .064950924579f, .016418869426f, 
				      -.225558972234f, -.100240215031f, .667071338154f, 
				      1.088251530500f, .542813011213f, -.050256540092f, 
				      -.045240772218f, .070703567550f, .008152816799f, 
				      -.028786231926f, -.001137535314f, .006495728375f, 
				      .000080661204f, -.000649589896f };
	       f = h;
	    }
	    break;
	 default :
	    std::cerr<<"This order is not implemented "<<std::endl;
	    Exit( FAILURE);
	 }
      } else if(!strcmp(wavelet, "vaidyanathan")){		
	 size = 24;
	 static Float h[25] = { 0.0f, 		-.000062906118f, .000343631905f, -.000453956620f, 
				-.000944897136f, .002843834547f,	.000708137504f, 
				-.008839103409f, .003153847056f, .019687215010f, 
				-.014853448005f, -.035470398607f, .038742619293f, 
				.055892523691f, -.077709750902f, -.083928884366f, 
				.131971661417f, .135084227129f, -.194450471766f, 
				-.263494802488f, .201612161775f, .635601059872f, 
				.572797793211f, .250184129505f, .045799334111f };
			
	 f = h; 
      }
      else if(!strcmp(wavelet, "battle")){
			
	 switch(order){
	 case  1 :
	    {
	       size = 23;
	       Float g[13] = { 0.0f, 	0.578163f, 0.280931f, -0.0488618f, 
			       -0.0367309f, 0.012003f, 0.00706442f, 
			       -0.00274588f, -0.00155701f, 0.000652922f, 
			       0.000361781f, -0.000158601f, -0.0000867523f };
	       static Float h[24];
	       int l=12,k=0;
		
	       for(k=l; k<=2*l-1;k++)
		  h[k]=g[k-l+1];
	       for(k=1; k<=l-1;k++)
		  h[l-k]=h[l+k];
	       f=h;
	    }
	    break;
	 case 3 :
	    {	size = 41;
	    Float g[22] = { 0.0f, 	0.541736f, 0.30683f, -0.035498f,	
			    -0.0778079f, 0.0226846f, 0.0297468f, 
			    -0.0121455f, -0.0127154f, 0.00614143f, 
			    0.00579932f, -0.00307863f, -0.00274529f, 
			    0.00154624f, 0.00133086f, -0.000780468f, 
			    -0.00065562f, 0.000395946f, 0.000326749f, 
			    -0.000201818f, -0.000164264f, 0.000103307f };
	    static Float h[42];
	    int l=21,k=0;
		
	    for(k=l; k<=2*l-1;k++)
	       h[k]=g[k-l+1];
	    for(k=1; k<=l-1;k++)
	       h[l-k]=h[l+k];
	    f=h;
	    }
	    break;
	 case 5 :
	    {
	       size = 59;
	       Float g[31] = { 0.0f, 	0.528374f, 0.312869f, -0.0261771f, 
			       -0.0914068f, 0.0208414f, 0.0433544f, 
			       -0.0148537f, -0.0229951f, 0.00990635f, 
			       0.0128754f, -0.00639886f, -0.00746848f, 
			       0.00407882f, 0.00444002f, -0.00258816f, 
			       -0.00268646f, 0.00164132f, 0.00164659f, 
			       -0.00104207f, -0.00101912f, 0.000662836f, 
			       0.000635563f, -0.000422485f, -0.000398759f, 
			       0.000269842f, 0.000251419f, -0.000172685f, 
			       -0.000159168f, 0.000110709f, 0.000101113f };
	       static Float h[60];
	       int l=30,k=0;
		
	       for(k=l; k<=2*l-1;k++)
		  h[k]=g[k-l+1];
	       for(k=1; k<=l-1;k++)
		  h[l-k]=h[l+k];
	       f=h;
	    }
	    break;
	 default :
	    std::cerr<<"This order is not implemented."<<std::endl;
	    Exit( FAILURE);
	 }
      }
      else if(!strcmp(wavelet, "haar")){
			
	 size = 2;
	 static Float h[3] = { 0.0f,0.70710678f, 0.70710678f };
	 f = h; 
      }
      else{
	 std::cerr<< wavelet <<" is not implemented. "<<std::endl;
	 Exit(FAILURE);
      }
		
      normf = norm(f,size);
      for(int i = 1; i <= size; i++)
	 f[i] = f[i]/normf;
		
      NameQmf =  wavelet;
      Size   = size +1;
      Coeffs = f;
		
      Save(*this);
	
   }
	
   /**
    * Gives the number of coefficients.
    */
   int GetSize() const   { return Size;   }
	
   /**
    * Gives the coefficients.
    */
   Float * GetCoeffs()   { return Coeffs; }
	
   /**
    * Gives the name of the QMF.
    */
   char  * GetName()     { return NameQmf;}
};

/**
 * This class defines the class Filter : 
 */
class Filter{

public :
	
   // Length of the filter
   int Size;
		
   // Origin of the filter           
   unsigned int Shift; 
		
   // Vector of coefficients  of low-pass analyse filter.
   Float *H;
		
   // Vector of coefficients  of high-pass analyse filter.
   Float *G;
		
   // Vector of coefficients of low-pass reconstruction filter.          
   Float *Hr;
		
   // Vector of coefficients of high-pass reconstruction filter.          
   Float *Gr;

   /**
    * Constructor
    */
   Filter():Size(0),Shift(0),H(0),G(0),Hr(0),Gr(0){}
	
   /**
    * Destructor
    */
   virtual ~Filter(){}
	
   /**
    * This function loads the filter which will be used by
    * the transform.
    * Inputs
    * 	q : The QMF.
    */
   virtual Errc LoadFilter(Qmf & q) = 0;
	
   /**
    * This function loads the filter which will be used by
    * the inverse transform.
    * Inputs
    * 	q : The QMF.
    */
   virtual Errc LoadRecFilter(Qmf &c) = 0;
};

/**
 * This class defines the filter which 
 * will be used by DWT and IDWT.
 */
class DwtFilter : public Filter{

public :

   /**
    * Destructor.
    */
   ~DwtFilter(){
      if(G)	delete []G;
   }
	
   /**
    * This function loads the filter used by DWT
    * Inputs
    * 	q : The QMF.
    */
   Errc LoadFilter(Qmf & q){
      int k;
      int s = -1;
	
      H    = q.GetCoeffs();
      Size = q.GetSize() - 1;
		
      G  = new Float[Size+1];
      G[0] = 0;
      for (k = 1; k <= Size; k++){
	 G[Size+1-k] = s*H[k];
	 s = -s;
      }
      Shift = -(Size >> 1);
      return SUCCESS;
	
   }
	
   /**
    * This function loads the filter used by IDWT
    * Inputs
    * 	q : The QMF.
    */
   Errc LoadRecFilter(Qmf &q){return LoadFilter(q);}
};

/**
 * This class defines the filter used by
 * UDWT and IUDWT
 */
class UdwtFilter : public Filter{

public :

   /**
    * Destructor
    */
   ~UdwtFilter(){
      if(H)  delete []H;
      if(G)  delete []G;
      if(Hr) delete []Hr;
      if(Gr) delete []Gr;
   }
	
   /**
    * This function loads the filter used by UDWT.
    * Inputs
    * 	q : The QMF.
    */
   Errc LoadFilter(Qmf & q){
      int i;
      int s = -1;
		
      Float* h    = q.GetCoeffs() + 1;
      Size        = q.GetSize()   - 1;
			
      H = new Float[Size];
      G = new Float[Size];
			
      /* analysis lowpass and highpass */
      for (i = 0; i < Size; i++){
	 H[i] = h[Size-i-1];
	 G[i] = s*h[i];
	 s = -s;
      }
      return SUCCESS;
   }
	
   /**
    * This function loads the filter used by the IUDWT.
    * Inputs
    * 	q : The QMF. 
    */
   Errc LoadRecFilter(Qmf &q){
      int i;
      Float* h    = q.GetCoeffs() +1;
      Size        = q.GetSize()   - 1;
		
      Hr = new Float[Size];
      Gr = new Float[Size];
		
      /* reconstruction lowpass and highpass */
      for (i = 0; i < Size; i++){
	 Hr[i] = h[i]/2;
	 Gr[i] = h[Size-i-1]/2;
      }
		
      for (i = 1; i < Size; i+=2)
	 Gr[i] = -Gr[i];
      return SUCCESS;
   }
};

#endif
